#include <Arduino.h>
#include <LedControl.h>

#define UP 1
#define RIGHT 2
#define DOWN 3
#define LEFT 4
#define CLK 10
#define CS 11
#define DIN 12
#define ANALOGSTICK_GND 14
#define ANALOGSTICK_VCC 15
#define ANALOGSTICK_X A2
#define ANALOGSTICK_Y A3
#define ACTIVE_BUZZER A4
#define POTENTIOMETER A5

struct Position {
    int x = 0, y = 0;
    Position() {}
    Position(int new_x, int new_y) {
        x = new_x;
        y = new_y;
    }
};

/* pozitiile sarpelui, a marului si cea default a Analog Stickului*/
Position snake, apple = {.x = -1, .y= -1}, analogStickDefaultPos = {.x = 500, .y= 500};

/* matricea corespunzatoare display-ului fizic */
LedControl ledMatrix(DIN, CLK, CS, 1);

/* intensitatea luminii de pe ecran */
const short ledIntensity = 5;

/* lungimea initiala a sarpelui */
const short initialSnakeLen = 3;

/* lungimea, viteza si directia initiala a sarpelui(0 - nu se misca) */
int snakeLen = initialSnakeLen, snakeSpeed = 1, snakeDir = 0;


/* varibile logice ale jocului */
bool isWon = false, isGameOver = false, isEaten;

/* pragul unde miscarea Analog Stickului va fi acceptata */
const int analogStickThreshold = 150;

/* matricea corespunzatoare pozitiilor de pe ecran */
int board[8][8];


/* functie de mapare folosita pentru calculul vitezei */
float speedMap(float x, float inMin, float inMax, float outMin, float outMax) {
    float xf = x - inMin;
    float out = outMax - outMin;
    float in = inMax - inMin;
    return xf * out / in + outMin;
}

/* daca sarpele depaseste o margine, se va proiecta in marginea opusa */
void edgePassing() {
    if (snake.y < 0) {
        snake.y += 8;
    }
    if (snake.y > 7) {
        snake.y -= 8;
    }
    if (snake.x < 0) {
        snake.x += 8;
    }
    if (snake.x > 7) {
        snake.x -= 8;
    }
}

void reloadGame() {
    /* ledul de pe pozitia marului se stinge */
    ledMatrix.setLed(0, apple.x, apple.y, 0);

    delay(800);

    /* se aprinde si stinge succesiv ecranul */
    for (int k = 0; k < 5; ++k) {
        for (int i = 0; i < 8; ++i) {
            for (int j = 0; j < 8; ++j) {
                if (board[i][j]) {
                    ledMatrix.setLed(0, i, j, 0);
                }
                else {
                    ledMatrix.setLed(0, i, j, 1);
                }
            }
        }

        delay(20);

        for (int i = 0; i < 8; ++i) {
            for (int j = 0; j < 8; ++j) {
                if (board[i][j]) {
                    ledMatrix.setLed(0, i, j, 1);
                }
                else {
                    ledMatrix.setLed(0, i, j, 0);
                }
            }
        }

        delay(50);
    }

    delay(600);

    for (int k = 1; k <= snakeLen; ++k) {
        for (int i = 0; i < 8; ++i) {
            for (int j = 0; j < 8; ++j) {
                if (board[i][j] == k) {
                    ledMatrix.setLed(0, i, j, 0);
                    delay(100);
                }
            }
        }
    }
}

/* functie care genereaza marul pe o pozitie de pe ecran
 * si verifica daca jocul e castigat
 */
void generateApple() {
    if (apple.x == -1 || apple.y == -1) {
        if (snakeLen >= 64) {
            isWon = true;
            return;
        }

        /* pozitia marului se genereaza pana cand nu se mai afla pe o pozitie din sarpe */
        do {
            apple.y = random(8);
            apple.x = random(8);
        } while (board[apple.x][apple.y] > 0);
    }
}

/* recepteaza miscarile din Analog Stick */
void scanAnalogStick() {
    int prevDir = snakeDir;
    unsigned long elapsedTime = millis();

    while (millis() < elapsedTime + snakeSpeed) {
        /* se calculeaza exponential viteza */
        float convertedSpeed = speedMap(analogRead(POTENTIOMETER), 0, 1000, 0, 1);
        snakeSpeed = speedMap(pow(convertedSpeed, 2.5), 0, 1, 10, 1000);
        if (snakeSpeed == 0) {
            snakeSpeed = 1;
        }

        /* citeste inputul din Analog stick si alege directia sarpelui */
        if (analogRead(ANALOGSTICK_Y) < analogStickDefaultPos.y - analogStickThreshold) {
            snakeDir = UP;
        }
        if (analogRead(ANALOGSTICK_Y) > analogStickDefaultPos.y + analogStickThreshold) {
            snakeDir = DOWN;
        }
        if (analogRead(ANALOGSTICK_X) < analogStickDefaultPos.x - analogStickThreshold) {
            snakeDir = LEFT;
        }
        if (analogRead(ANALOGSTICK_X) > analogStickDefaultPos.x + analogStickThreshold) {
            snakeDir = RIGHT;
        }


        /* se evita coliziunea in cazul in care urmatorul input va fi la 180 de grade
         * fata de sensul precedent
         */
        if ((snakeDir + 2 == prevDir || snakeDir - 2 == prevDir) && prevDir != 0 ) {
            snakeDir = prevDir;
        }

        /* ledul corespunzator marului se aprinde si stinge la cateva milisecunde */
        if (millis() % 100 < 50) {
            ledMatrix.setLed(0, apple.x, apple.y, 1);
        } else {
            ledMatrix.setLed(0, apple.x, apple.y, 0);
        }
    }
}

// actualizeaza pozitia sarpelui si verifica daca jocul s-a terminat 
void computeSnakeAttributes() {
    bool ok = false;

    if (snakeDir == UP) {
        snake.x--;
        edgePassing();
        ok = true;
    } else if (snakeDir == DOWN) {
        snake.x++;
        edgePassing();
        ok = true;
    } else if (snakeDir == LEFT) {
        snake.y--;
        edgePassing();
        ok = true;
    } else if (snakeDir == RIGHT) {
        snake.y++;
        edgePassing();
        ok = true;
    }

    if (ok) {
        ledMatrix.setLed(0, snake.x, snake.y, 1);
    } else {
        return;
    }

    /* jocul se termina daca pozitia din varful sarpelui
     * (conventie: aceasta nu face parte din segment)
     * ajunge pe o pozitie din segment, el aflandu-se in miscare
     */
    if (board[snake.x][snake.y] > 1 && snakeDir != 0) {
        isGameOver = true;
        return;
    }

    /* se verifica daca sarpele a mancat marul si se actualizeaza lungimea si pozitiile corpului
     * se creste valoarea de pe fiecare pozitie cu 1
     */
    if (snake.x == apple.x && snake.y == apple.y) {
        apple.x = -1;
        apple.y = -1;
        snakeLen++;
        for (int i = 0; i < 8; ++i) {
            for (int j = 0; j < 8; ++j) {
                if (board[i][j] > 0) {
                    board[i][j]++;
                }
            }
        }
        isEaten = true;
    }

    /* se adauga o unitate la lungimea sarpelui in varful acestuia */
    board[snake.x][snake.y] = snakeLen + 1;

    /* se scade cu 1 valoarea de pe fiecare pozitie a sarpelui, iar, daca aceasta este mai mica de 1,
     * se stinge ledul ce corespunde pozitiei respective
     */
    for (int i = 0; i < 8; ++i) {
        for (int j = 0; j < 8; ++j) {
            if (board[i][j] > 0 ) {
                board[i][j]--;
            }

            if (board[i][j]) {
                ledMatrix.setLed(0, i, j, 1);
            } else {
                ledMatrix.setLed(0, i, j, 0);
            }
        }
    }
}

/* calibreaza pozitia default a Analog Stickului in urma fortelor
 * de impingere si revenire exercitate asupra sa
 */
void calibrateAnalogStick() {
    Position inputs;

    for (int i = 0; i < 10; i++) {
        inputs.x += analogRead(ANALOGSTICK_X);
        inputs.y += analogRead(ANALOGSTICK_Y);
    }

    analogStickDefaultPos.x = inputs.x / 10;
    analogStickDefaultPos.y = inputs.y / 10;
}

/* buzzerul se activeaza timp de 0.1 sec dupa ce s-a mancat marul */
void activateBuzzer() {
    if (isEaten) {
        digitalWrite(ACTIVE_BUZZER, HIGH);
        delay(100);
        digitalWrite(ACTIVE_BUZZER, LOW);
    }
    isEaten = false;
}

/* initializarea, pinilor, ledurilor matricei si a pozitiei sarpelui */
void initPinsAndLeds() {
    pinMode(ANALOGSTICK_VCC, OUTPUT);
    pinMode(ANALOGSTICK_GND, OUTPUT);
    pinMode(ACTIVE_BUZZER, OUTPUT);
    digitalWrite(ANALOGSTICK_VCC, HIGH);
    digitalWrite(ANALOGSTICK_GND, LOW);

    ledMatrix.setIntensity(0, ledIntensity);
    ledMatrix.shutdown(0, false);
    ledMatrix.clearDisplay(0);

    randomSeed(analogRead(A5));
    snake.x = random(8);
    snake.y = random(8);
}

/* reinitializarea datelor jocului */
void reconfigureGameData() {
    if (isGameOver || isWon) {
        reloadGame();
        isWon = false;
        isGameOver = false;
        snakeLen = initialSnakeLen;
        snakeDir = 0;
        snake.x = random(8);
        snake.y = random(8);
        apple.x = -1;
        apple.y = -1;
        memset(board, 0, sizeof(board[0][0]) * 8 * 8);
        ledMatrix.clearDisplay(0);
    }
}

/* afisarea scorului */
void printScore(){
    Serial.print("Score: ");
    Serial.println(snakeLen - initialSnakeLen);
}

void setup() {
    Serial.begin(115200);
    initPinsAndLeds();
    calibrateAnalogStick();
}

void loop() {
    printScore();
    generateApple();
    scanAnalogStick();
    computeSnakeAttributes();
    activateBuzzer();
    reconfigureGameData();
}
